package data_protocol

import (
	"encoding/json"
	"git.sxidc.com/go-tools/utils/strutils"
	"github.com/pkg/errors"
	"time"
)

const (
	DataContentTypeJson = "application/json"
)

// NewCloudEvent 创建CloudEvent
// 参数:
// - eventID: 事件ID
// - eventType: 事件类型
// - source: 事件源，如foo.com
// - dataContentType: 事件数据类型，可以是任何mime类型，如application/json，application/text
// - data: 要发送的数据
// 返回值:
// - CloudEvent
func NewCloudEvent(eventID string, eventType string, source string, dataContentType string, data []byte) *CloudEvent {
	return &CloudEvent{
		SpecVersion:     "v1.0",
		ID:              eventID,
		Type:            eventType,
		Source:          source,
		DataContentType: dataContentType,
		Time:            time.Now().Format(time.RFC3339),
		Data:            data,
	}
}

// NewCloudEventJson 创建数据是Json类型的CloudEvent
// 参数:
// - eventID: 事件ID
// - eventType: 事件类型
// - source: 事件源，如foo.com
// - obj: 需要转为json的对象
// 返回值:
// - CloudEvent
// - 错误
func NewCloudEventJson(eventID string, eventType string, source string, obj any) (*CloudEvent, error) {
	dataJsonBytes, err := json.Marshal(obj)
	if err != nil {
		return nil, errors.New(err.Error())
	}

	return &CloudEvent{
		SpecVersion:     "v1.0",
		ID:              eventID,
		Type:            eventType,
		Source:          source,
		DataContentType: DataContentTypeJson,
		Time:            time.Now().Format(time.RFC3339),
		Data:            dataJsonBytes,
	}, nil
}

// UnmarshalJsonCloudEvent Unmarshal格式为Json的CloudEvent
// 参数:
// - cloudEventJsonBytes: CloudEvent的Json字节
// 返回值:
// - CloudEvent
// - 错误
func UnmarshalJsonCloudEvent(cloudEventJsonBytes []byte) (*CloudEvent, error) {
	event := new(CloudEvent)
	err := json.Unmarshal(cloudEventJsonBytes, event)
	if err != nil {
		return nil, errors.New(err.Error())
	}

	return event, nil
}

type CloudEvent struct {
	SpecVersion     string `json:"specversion"`
	ID              string `json:"id"`
	Type            string `json:"type"`
	Source          string `json:"source"`
	DataContentType string `json:"datacontenttype"`
	Time            string `json:"time"`
	Data            []byte `json:"data"`
}

// MarshalJson Marshal格式为Json的CloudEvent
// 参数: 无
// 返回值:
// - CloudEvent的Json字节
// - 错误
func (event *CloudEvent) MarshalJson() ([]byte, error) {
	if strutils.IsStringEmpty(event.SpecVersion) {
		return nil, errors.New("没有传递事件规范版本")
	}

	if strutils.IsStringEmpty(event.ID) {
		return nil, errors.New("没有传递事件ID")
	}

	if strutils.IsStringEmpty(event.Type) {
		return nil, errors.New("没有传递事件类型")
	}

	if strutils.IsStringEmpty(event.Source) {
		return nil, errors.New("没有传递事件源")
	}

	if strutils.IsStringEmpty(event.DataContentType) {
		return nil, errors.New("没有传递事件数据类型")
	}

	if strutils.IsStringEmpty(event.Time) {
		return nil, errors.New("没有传递事件时间")
	}

	eventJson, err := json.Marshal(event)
	if err != nil {
		return nil, errors.New(err.Error())
	}

	return eventJson, nil
}
