package local

import (
	"github.com/pkg/errors"
	"strings"
	"sync"
	"time"
)

type valueItem struct {
	value           string
	expireStartTime time.Time
	expireSec       int64
	expireDoneChan  chan any
}

type Cache struct {
	*sync.RWMutex
	namespace string
	cacheMap  map[string]*valueItem
}

func New(namespace string) *Cache {
	return &Cache{
		RWMutex:   new(sync.RWMutex),
		namespace: namespace,
		cacheMap:  make(map[string]*valueItem),
	}
}

func Destroy(cache *Cache) {
	if cache == nil {
		return
	}

	cache.Lock()
	defer cache.Unlock()

	for _, item := range cache.cacheMap {
		if item.expireDoneChan != nil {
			item.expireDoneChan <- nil
		}
	}

	cache.cacheMap = nil
}

func (cache *Cache) Set(key string, value string, expireSec int64) error {
	cache.Lock()
	defer cache.Unlock()

	oldItem, ok := cache.cacheMap[cache.key2CacheKey(key)]
	if ok {
		cache.stopExpireRoutine(oldItem)

		newItem := &valueItem{
			value:           value,
			expireStartTime: time.Now(),
			expireSec:       expireSec,
		}

		cache.cacheMap[cache.key2CacheKey(key)] = newItem

		if expireSec == 0 {
			return nil
		}

		cache.startExpireRoutine(key, newItem)

		return nil
	}

	item := &valueItem{
		value:           value,
		expireStartTime: time.Now(),
		expireSec:       expireSec,
	}

	cache.cacheMap[cache.key2CacheKey(key)] = item

	if item.expireSec != 0 {
		cache.startExpireRoutine(key, item)
	}

	return nil
}

func (cache *Cache) Get(key string) (string, error) {
	cache.RLock()
	defer cache.RUnlock()

	item, ok := cache.cacheMap[cache.key2CacheKey(key)]
	if !ok {
		return "", nil
	}

	return item.value, nil
}

func (cache *Cache) GetMulti(keys []string) (map[string]string, error) {
	cache.RLock()
	defer cache.RUnlock()

	result := make(map[string]string)

	for _, key := range keys {
		item, ok := cache.cacheMap[cache.key2CacheKey(key)]
		if !ok {
			result[key] = ""
		}

		result[key] = item.value
	}

	return result, nil
}

func (cache *Cache) GetAll() (map[string]string, error) {
	cache.RLock()
	defer cache.RUnlock()

	result := make(map[string]string)

	for key, item := range cache.cacheMap {
		result[cache.cacheKey2Key(key)] = item.value
	}

	return result, nil
}

func (cache *Cache) Delete(key string) error {
	cache.Lock()
	defer cache.Unlock()

	delete(cache.cacheMap, cache.key2CacheKey(key))
	return nil
}

func (cache *Cache) Clear() error {
	cache.Lock()
	defer cache.Unlock()

	cache.cacheMap = make(map[string]*valueItem)
	return nil
}

func (cache *Cache) Expire(key string, expireSec int64) error {
	cache.Lock()
	defer cache.Unlock()

	oldItem, ok := cache.cacheMap[cache.key2CacheKey(key)]
	if !ok {
		return errors.New("对应的键不存在: " + key)
	}

	cache.stopExpireRoutine(oldItem)

	newItem := &valueItem{
		value:           oldItem.value,
		expireStartTime: time.Now(),
		expireSec:       expireSec,
	}

	cache.cacheMap[cache.key2CacheKey(key)] = newItem

	if expireSec == 0 {
		return nil
	}

	cache.startExpireRoutine(key, newItem)

	return nil
}

func (cache *Cache) TTL(key string) (int64, error) {
	cache.RLock()
	defer cache.RUnlock()

	item, ok := cache.cacheMap[cache.key2CacheKey(key)]
	if !ok {
		return 0, nil
	}

	remainSec := int64(item.expireStartTime.Add(time.Duration(item.expireSec)*time.Second).Sub(time.Now()).Seconds() + 0.5)
	if remainSec <= 0 {
		return 0, nil
	}

	return remainSec, nil
}

func (cache *Cache) startExpireRoutine(key string, item *valueItem) {
	item.expireDoneChan = make(chan any)

	go func() {
		timer := time.NewTimer(time.Second * time.Duration(item.expireSec))

		defer func() {
			if item.expireDoneChan != nil {
				close(item.expireDoneChan)
				item.expireDoneChan = nil
			}
		}()

		for {
			select {
			case <-timer.C:
				delete(cache.cacheMap, cache.key2CacheKey(key))
				return
			case <-item.expireDoneChan:
				timer.Stop()
				return
			}
		}
	}()
}

func (cache *Cache) stopExpireRoutine(item *valueItem) {
	if item.expireDoneChan != nil {
		item.expireDoneChan <- nil
	}
}

func (cache *Cache) key2CacheKey(key string) string {
	return cache.namespace + "::" + key
}

func (cache *Cache) cacheKey2Key(cacheKey string) string {
	return strings.TrimPrefix(cacheKey, cache.namespace+"::")
}
